(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory ProofChecker imports Substitution RedPair DepGraph Polynomial DpTransImpl Loop begin



text {* 
   A polynomial order is represented by a list of function symbol assignments where 
   every function symbol obtains a pair of a number n and a list of numbers [n1,..,nk].
   The corresponding polynomial interpretation is as follows:
   Pol(f(x1,..,xl)) = n + sum_i ni * xi
        where i ranges from 1 to min(k,l)
*}

datatype ('f,'v)"DPProof" = PisEmpty  
                          | RedPairProc "'f RedPair" "('f,'v)trsL"  "('f,'v) DPProof"
                          | DepGraphProc "(('f,'v)DPProof option \<times> ('f,'v)trsL) list"

text {* appropriate wrappers for loop *}
types ('f,'v) subL = "(('f,'v)term \<times>('f,'v)term) list"
types ('f,'v) loopL = "('f,'v)term list \<times> ('f,'v)ctxt \<times> ('f,'v)subL"

fun "checkLoop" :: "('f :: sshow,'v :: sshow)trsL \<Rightarrow> ('f,'v)loopL \<Rightarrow> sstring list monad" where
 "checkLoop R (ts,C,sub) = ((case Loop.subst_safe [] sub \<iota> of None \<Rightarrow> error ''substitution not unique''
                                                           | Some \<sigma> \<Rightarrow> check_loop (ts,C,\<sigma>) R)
      \<leftarrow> (\<lambda>s. [''error in certification of loop'',s]))"


datatype ('f,'v)"TRSProof" = DPTrans "('f,'v)dpL"  "('f shp, 'v) DPProof"
                           | Loop "('f,'v)loopL"


fun indent :: "sstring list \<Rightarrow> sstring list"
where "indent xs = (newline # (map (\<lambda> s. ''   '' @ s) xs))" 


fun checkDPProof :: "('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f :: sshow, 'v :: sshow)DPProof \<Rightarrow> sstring list monad" 
where 
   "checkDPProof P R PisEmpty = 
            (case P of Nil \<Rightarrow> return () | Cons _ _ \<Rightarrow> 
                     error (''P is not empty in the following DP-problem '' # showDPP P R))"  
|  "checkDPProof P R (RedPairProc redp rP prf)  = (
         createRedPairProc redp P rP R 
           \<leftarrow> (\<lambda> s. showDPP P R @ (''error when applying the reduction pair processor to obtain'' # (showP rP @ s))) 
       \<guillemotright> checkDPProof rP R prf
           \<leftarrow> (\<lambda> s. ''error below the reduction pair processor'' # (indent s))
)" 
| "checkDPProof P R (DepGraphProc edpts) = (
       checkDepGraphProc P (map (\<lambda> x. ((case fst x of Some _ \<Rightarrow> True | None \<Rightarrow> False), snd x)) edpts) R
           \<leftarrow> (\<lambda> s. ''error while trying to perform SCC-decomposition  on '' # 
                          (showDPP P R @  s))
     \<guillemotright> checkAllM (\<lambda> rp. case fst rp of Some prf \<Rightarrow> checkDPProof (snd rp) R prf | None \<Rightarrow> OK) edpts
           \<leftarrow> (\<lambda> s. ''error below the dependency graph processor'' # (indent s))
)
" 


lemma checkDPProofSound[simp]  : fixes P :: "('f :: sshow,'v :: sshow)trsL"
  assumes "isOK(checkDPProof P R prf)"
  shows "finite_dpp (set P, set R)"
using assms
proof (induct rule:checkDPProof.induct)
  case (1 P R) 
  thus "finite_dpp (set P, set R)" unfolding finite_dpp_def by (cases P, auto)
next
  case (2 P R redp rP prof) 
  thus ?case by auto
next
  case (3 P R edpts) 
  let ?mc = "\<lambda> edpts. (map (\<lambda> x. ((case fst x of Some _ \<Rightarrow> True | None \<Rightarrow> False), snd x)) edpts)"
  let ?edps = "?mc edpts"
  show ?case 
  proof (rule checkDepGraphProcSound, auto)
    from 3 show "isOK(checkDepGraphProc P ?edps R)" by (auto)
  next
    fix a rp
    assume elem: "(True, rp) \<in> set ?edps"
    from 3 have fin: "\<forall> prf rP. (Some prf, rP) \<in> set edpts \<longrightarrow> finite_dpp (set rP, set R)" by auto
    from elem obtain sub where "set sub \<subseteq> set edpts \<and> (True, rp) \<in> set (?mc sub)" by blast
    thus "finite_dpp (set rp, set R)"
    proof (induct sub, simp)
      case (Cons edp subb)
      with fin show ?case 
	by (cases "edp", cases "fst edp", simp, cases "snd edp = rp", auto)
    qed
  qed
qed


declare checkDPProof.simps[simp del]


fun checkTRSProof :: "('f :: sshow, 'v :: sshow)trsL \<Rightarrow> ('f,'v) TRSProof \<Rightarrow> sstring list monad"
where "checkTRSProof r (DPTrans P prf) =  (
    checkDPs r P
      \<leftarrow> (\<lambda> s. ''error when switching from the TRS'' # 
                 ( showTRS r @  (''to the initial DP problem'' # showDPP P (computeID r) @  s)))
  \<guillemotright> checkDPProof P (computeID r) prf
      \<leftarrow> (\<lambda> s. ''error below switch to dependency pairs'' # (indent s))
)"
  | "checkTRSProof r (Loop l) = checkLoop r l"

(*
lemma checkTRSProofSound  : 
  assumes check: "isOK(checkTRSProof r prof)"
  shows "SN (rstep (set r))"
using assms
proof (cases "prof", auto)
  fix P dpprf
  assume dpsSound: "isOK(checkDPs r P)" and rec: "isOK(checkDPProof P (computeID r) dpprf)"
  from rec have "finite_dpp (set P, set (computeID r))" by (rule checkDPProofSound)
  with dpsSound show ?thesis by (simp)
qed
*)

lemma checkTRSProofSound  : 
  assumes check: "isOK(checkTRSProof r (DPTrans dp dpprof))"
  shows "SN (rstep (set r))"
using assms
proof (auto)
  fix P dpprf
  assume dpsSound: "isOK(checkDPs r P)" and rec: "isOK(checkDPProof P (computeID r) dpprf)"
  from rec have "finite_dpp (set P, set (computeID r))" by (rule checkDPProofSound)
  with dpsSound show ?thesis by (simp)
qed

lemma checkTRSProofSound2 :
 assumes "isOK(checkTRSProof r (Loop (ts,C,sub)))"
 shows "\<not>SN (rstep (set r))"
using assms by (cases "subst_safe [] sub \<iota>") auto 

end
